// synteny : checking of the predicate
#include "synteny.h"

using namespace std;

// Return whether 1st contig contains more genes than 2nd contig
bool    GreaterGenes(Citer c1, Citer c2) {
    return c1->second.genes.size() > c2->second.genes.size();
}

// subrange: 0=entire vicinity, -1=left incl. X, +1=right incl. X, -2/+2=...excl. X
bool  ExceedSpan(const Group & grp, const int span, const int subrange) {
    if (grp.gsize == 1) return false;
    else if (grp.gsize == 2 || grp.gsize == 3 && !grp.pgene[2]) {
        if (!grp.pgene[0] || !grp.pgene[1])
            return true;    //false;
        Gene & x = *grp.pgene[0];
        Gene & y = *grp.pgene[1];

        if (grp.pgene[0] == grp.pgene[1])
            return true;   // Skip degenerated pairs
        if (subrange == -1 && y.end > x.end || subrange == -2 && y.end >= x.begin)
            return true;   // Skip violating left subrange condition
        if (subrange == 1 && y.begin < x.begin || subrange == 2 && y.begin <= x.end)
            return true;   // Skip violating right subrange condition

        if (global.span_outer)
            return __max(x.end, y.end) - __min(x.begin, y.begin) > span;
        else if (global.span_center)
            return abs(x.begin + x.end - y.begin - y.end) > span*2;
        else 
            return (x.begin < y.end ? y.begin - x.end : x.begin - y.end) > span;
    }
    else if (grp.gsize == 3) {
        if (!grp.pgene[0] || !grp.pgene[1])
            return true;    //false;
        Gene & x = *grp.pgene[0];
        Gene & y = *grp.pgene[1];
        Gene & z = *grp.pgene[2];

        if (grp.pgene[0] == grp.pgene[1] || grp.pgene[0] == grp.pgene[2] || 
            ZneY && grp.pgene[1] == grp.pgene[2])
                return true;   // Skip degenerated triplets
        if (subrange == -1 && (y.end > x.end || z.end > x.end) || 
            subrange == -2 && (y.end >= x.begin || z.end >= x.begin))
            return true;   // Skip violating left subrange condition
        if (subrange == 1 && (y.begin < x.begin || z.begin < x.begin) || 
            subrange == 2 && (y.begin <= x.end || z.begin <= x.end))
            return true;   // Skip violating right subrange condition

        if (global.span_outer) {
            int rxy = __max(x.end, y.end) - __min(x.begin, y.begin);
            int rxz = __max(x.end, z.end) - __min(x.begin, z.begin);
            return __max(rxy, rxz) > span;
        }
        else if (global.span_center) {
            int rxy = abs(x.begin + x.end - y.begin - y.end);
            int rxz = abs(x.begin + x.end - z.begin - z.end);
            return __max(rxy, rxz) > span*2;
        }
        else {
            int rxy = x.begin < y.end ? y.begin - x.end : x.begin - y.end;
            int rxz = x.begin < z.end ? z.begin - x.end : x.begin - z.end;
            return __max(rxy, rxz) > span;
        }
    }
    return false;
}

bool  InSynteny(const Group & ga, const Group & gb) {

    if (ga.gsize != gb.gsize) return false;

    if (ga.gsize == 1)
        return (ga.pgene[0] != 0 && gb.pgene[0] != 0);

    else if (ga.gsize == 2) {
        if (ga.pgene[0]==0 || gb.pgene[0]==0 || ga.pgene[1]==0 || gb.pgene[1]==0)
            return false;
        Gene & x    = *(ga.pgene[0]);
        Gene & y    = *(ga.pgene[1]);
        Gene & x1   = *(gb.pgene[0]);
        Gene & y1   = *(gb.pgene[1]);

        if (global.mode_direct) {
            if (x.strand == x1.strand && y.strand == y1.strand || 
                global.mode_mirror && x.strand == -x1.strand && y.strand == -y1.strand)
                ;     // fall through
            else
                return false;
        }

        if (global.mode_order && !global.mode_direct) {
            if (x.begin < y.begin && x1.begin < y1.begin || 
                x.begin >= y.begin && x1.begin >= y1.begin ||
                global.mode_mirror && x.begin < y.begin && x1.begin >= y1.begin ||
                global.mode_mirror && x.begin >= y.begin && x1.begin < y1.begin)
                ;     // fall through
            else
                return false;
        }
        else if (global.mode_order) {
            if (x.strand == x1.strand && x.begin < y.begin && x1.begin < y1.begin ||
                x.strand == x1.strand && x.begin >= y.begin && x1.begin >= y1.begin ||
                global.mode_mirror && x.strand != x1.strand && x.begin < y.begin && x1.begin >= y1.begin ||
                global.mode_mirror && x.strand != x1.strand && x.begin >= y.begin && x1.begin < y1.begin)
                ;     // fall through
            else
                return false;
        }

        return true;
    }

    else if (ga.gsize == 3) {
        if (ga.pgene[0]==0 || ga.pgene[1]==0 || ga.pgene[2]==0 || 
            gb.pgene[0]==0 || gb.pgene[1]==0 || gb.pgene[2]==0)
                return false;
        Gene & x    = *(ga.pgene[0]);
        Gene & y    = *(ga.pgene[1]);
        Gene & z    = *(ga.pgene[2]);
        Gene & x1   = *(gb.pgene[0]);
        Gene & y1   = *(gb.pgene[1]);
        Gene & z1   = *(gb.pgene[2]);

        if (global.mode_direct) {
            if (x.strand == x1.strand && y.strand == y1.strand && z.strand == z1.strand || 
                global.mode_mirror && x.strand == -x1.strand && y.strand == -y1.strand && z.strand == -z1.strand)
                ;     // fall through
            else
                return false;
        }

        if (global.mode_order && !global.mode_direct) {
            if (x.begin < y.begin && y.begin < z.begin && x1.begin < y1.begin && y1.begin < z1.begin || 
                x.begin >= y.begin && y.begin >= z.begin && x1.begin >= y1.begin && y1.begin >= z1.begin ||
                x.begin < z.begin && z.begin < y.begin && x1.begin < z1.begin && z1.begin < y1.begin ||
                x.begin >= z.begin && z.begin >= y.begin && x1.begin >= z1.begin && z1.begin >= y1.begin ||
                y.begin < x.begin && x.begin < z.begin && y1.begin < x1.begin && x1.begin < z1.begin ||
                y.begin >= x.begin && x.begin >= z.begin && y1.begin >= x1.begin && x1.begin >= z1.begin ||
                global.mode_mirror && 
                (x.begin < y.begin && y.begin < z.begin && x1.begin >= y1.begin && y1.begin >= z1.begin ||
                x.begin >= y.begin && y.begin >= z.begin && x1.begin < y1.begin && y1.begin < z1.begin ||
                x.begin < z.begin && z.begin < y.begin && x1.begin >= z1.begin && z1.begin >= y1.begin ||
                x.begin >= z.begin && z.begin >= y.begin && x1.begin < z1.begin && z1.begin < y1.begin ||
                y.begin < x.begin && x.begin < z.begin && y1.begin >= x1.begin && x1.begin >= z1.begin ||
                y.begin >= x.begin && x.begin >= z.begin && y1.begin < x1.begin && x1.begin < z1.begin))
                ;     // fall through
            else
                return false;
        }
        else if (global.mode_order) {  // (&& global.mode_direct)
            if (x.strand == x1.strand && 
                (x.begin < y.begin && y.begin < z.begin && x1.begin < y1.begin && y1.begin < z1.begin ||
                x.begin >= y.begin && y.begin >= z.begin && x1.begin >= y1.begin && y1.begin >= z1.begin ||
                x.begin < z.begin && z.begin < y.begin && x1.begin < z1.begin && z1.begin < y1.begin ||
                x.begin >= z.begin && z.begin >= y.begin && x1.begin >= z1.begin && z1.begin >= y1.begin ||
                y.begin < x.begin && x.begin < z.begin && y1.begin < x1.begin && x1.begin < z1.begin ||
                y.begin >= x.begin && x.begin >= z.begin && y1.begin >= x1.begin && x1.begin >= z1.begin) ||
                global.mode_mirror && x.strand == -x1.strand && 
                (x.begin < y.begin && y.begin < z.begin && x1.begin >= y1.begin && y1.begin >= z1.begin ||
                x.begin >= y.begin && y.begin >= z.begin && x1.begin < y1.begin && y1.begin < z1.begin ||
                x.begin < z.begin && z.begin < y.begin && x1.begin >= z1.begin && z1.begin >= y1.begin ||
                x.begin >= z.begin && z.begin >= y.begin && x1.begin < z1.begin && z1.begin < y1.begin ||
                y.begin < x.begin && x.begin < z.begin && y1.begin >= x1.begin && x1.begin >= z1.begin ||
                y.begin >= x.begin && x.begin >= z.begin && y1.begin < x1.begin && x1.begin < z1.begin))
                ;     // fall through
                else
                    return false;
        }

        return true;
    }
    return false;
}

// Common variables for Predicate, HasSynteny, and CheckBack:
    Group g0, g1, g2, g3;       // groups for current predicate term (two pairs max)
    int tno = 0;                // current predicate term no.

// Return: satisfied/unsatisfied, side effects in groups and detail
bool Predicate(Gene & x, vector<Group> & groups) {
    // save previous solution for case of nonzero global.contin
    vector<Group> oldgroups(groups);
    vector<Group>::iterator ogit = oldgroups.begin();    // old group iterator
    groups.clear();

    g0.clear(); g0.pgene[0] = &x;
    g1.clear(); 
    g2.clear();
    g3.clear();
    int sp0=0, sp1, sp2, sp3;       // species/taxa numbers for current predicate term 
    sp1 = unknown_species;
    sp3 = unknown_species;
    Term & gt0 = global.terms[0];
    Term parminit = gt0;
    int range0 = gt0.range;
    int range1;
    int subrange = 0;
    int orcount  = -1;  // means no OR entered
    int andcount = -1;  // means no AND entered
    int overloop = 0;   // OVER-loop variables
    Giter oyit, ozit;
    bool exspan = false;
    bool logging = (global.mpisize == 1) && (global.log_verbose || x.log);
    bool logany = global.log_contig || global.log_contigno || global.log_gene ||
        global.log_geneno || global.log_protein || global.log_verbose;
    int y_no = 0;
    if (logging)
        cerr << "\nX=" << x.name << endl;

    for (tno = 0; tno >= 0 && tno < (int)global.terms.size(); ) {
        Term & t = global.terms[tno];

        if (logging) {
            cerr << "#" << tno << " " << TermOper[t.oper];
            if (t.oper == OVER || t.oper == OVERBK) 
                cerr << "_" << overloop;
            cerr << " \t";
            if (t.sp == unknown_species)
                cerr << setw(16) << " ";
            else if (t.sp >= 0 && t.sp < global.nAnimals)
                cerr << left << setw(16) << global.Animals[t.sp].name;
            else if (t.sp < 0 && t.sp > -(int)global.taxa.size())
                cerr << left << setw(16) << global.taxa[-t.sp].name;
            else
                cerr << left << setw(16) << "????????";
            if (t.yes == YES)
                cerr << right << setw(4) << label_yes;
            else if (t.yes == NO)
                cerr << right << setw(4) << label_no;
            else
                cerr << right << setw(4) << t.yes;
            if (t.no == t.yes)
                cerr << setw(4) << " ";
            else if (t.no == YES)
                cerr << right << setw(4) << label_yes;
            else if (t.no == NO)
                cerr << right << setw(4) << label_no;
            else
                cerr << right << setw(4) << t.no;
            if (t.oper==SET || t.oper==IN || t.oper==LIN || t.oper==LLIN || t.oper==RIN || 
                t.oper==RRIN || t.oper==INBK || t.oper==LINBK || t.oper==LLINBK || t.oper==RINBK || 
                t.oper==RRINBK || t.oper==IN2 || t.oper==OVER || t.oper==IN2BK || t.oper==OVERBK)
                cerr << "  " << right << setw(2) << t.nwit << " " << t.range << setw(3) 
					<< t.BBH << fixed << setprecision(3) << t.rank << setw(6) 
                    << setprecision(3) << t.score << scientific << setprecision(1) 
                    << setw(9) << t.E << fixed << setprecision(3) << " " << t.lambda 
                    << setw(3) << t.uselast0 << setw(3) << t.uselast1;
            else if (t.oper==ADD) {
                cerr << "   Fields:";
                bool first = true;
                for (int fo = 1; fo <= maxfield; fo++) {
                    int my = 0;
                    for ( ; my < maxfield; my++)
                        if (t.forder[my] == fo) {
                            cerr << (first? ' ': ',') << global.myfields[my];
                            first = false;
                            break;
                        }
                    if (my >= maxfield) break;
                }
            }
            cerr << endl;
        }

        if (t.oper == NOP || t.oper == GOTO || t.oper == NEXT) {
            tno = t.yes;
        }
        else if (t.oper == SET) {
            gt0.nwit    = t.nwit;
            gt0.range   = t.range;
            gt0.BBH     = t.BBH;
			gt0.rank	= t.rank;
			gt0.score	= t.score;
            gt0.E       = t.E;
            global.lambda = t.lambda;
            gt0.uselast0 = t.uselast0;
            gt0.uselast1 = t.uselast1;
            tno = t.yes;
        }
        else if (t.oper == BOR) {
            orcount = 0;
            if (logging) 
                cerr << "OR_count=" << orcount << endl;
            tno = t.yes;
        }
        else if (t.oper == EOR) {
            if (logging) 
                cerr << "OR_count=" << orcount << endl;
            tno = (orcount >= t.nwit ? t.yes : t.no);   // check number of positives (>=)
            orcount = -1;
        }
        else if (t.oper == EOS) {
            if (logging) 
                cerr << "OR_count=" << orcount << endl;
            tno = (orcount > t.nwit ? t.yes : t.no);   // check number of positives (>)
            orcount = -1;
        }
        else if (t.oper == BAND) {
            andcount = 0;
            if (logging) 
                cerr << "AND_count=" << andcount << endl;
            tno = t.yes;
        }
        else if (t.oper == EAND) {
            if (logging) 
                cerr << "AND_count=" << andcount << endl;
            tno = (andcount >= t.nwit ? t.yes : t.no);   // check number of negatives
            andcount = -1;
        }
        else if (t.oper == IN || t.oper == LIN || t.oper == LLIN || t.oper == RIN || t.oper == RRIN ||
                 t.oper == INBK || t.oper == LINBK || t.oper == LLINBK || t.oper == RINBK || t.oper == RRINBK) {
            switch (t.oper) {
                case IN:    case INBK:    subrange = 0;  break;
                case LIN:   case LINBK:   subrange = -1; break;
                case LLIN:  case LLINBK:  subrange = -2; break;
                case RIN:   case RINBK:   subrange = 1;  break;
                case RRIN:  case RRINBK:  subrange = 2;  break;
            }
            // Set HasSynteny mode for xIN/xINBK
            int hsmode = t.oper==INBK||t.oper==LINBK||t.oper==LLINBK||t.oper==RINBK||t.oper==RRINBK ? 1 : 0;
            bool insynt = false;
            int incase = 0;     // synteny case no. for logging
            for (int n=0; n<maxgroup; n++) {
                if ((t.uselast0 & (1<<n)) == 0)
                    g0.pgene[n] = 0;
            }
            g0.gsize = t.nwit + 1;
            g0.pgene[0] = &x;
            g0.goon = 0;
            SetBestHit(g0, 0);       // set E & ?BH for X
            if (logging) {
                cerr << "Group0[" << g0.gsize << "]";
                for (int k = 0; k < g0.gsize; k++)
                    cerr << " " << (g0.pgene[k]==0 ? "0" : g0.pgene[k]->name);
                cerr << endl;
            }
            sp1 = t.sp;
            g1.clear();
            g1.gsize = t.nwit + 1;
            if ((global.contin & 1) != 0 && ogit != oldgroups.end()) {   // continue from last X'
                g1.pgene[0] = (ogit + 1)->pgene[0];
                g1.goon = 1;
            }
            // Single loop until satisfied
            if (g0.gsize > 1) {     // Y required
                if (g0.pgene[1] == 0) {     // Select gene Y
                    Giter yit = x.cit->second.genes.begin();
                    for ( ; yit != x.cit->second.genes.end(); yit++) {
                        if (&(yit->second) == g0.pgene[0]) continue;
                        g0.pgene[1] = &(yit->second);
                        exspan = ExceedSpan(g0, range0, 0);    // Check span
                        if (logging && g0.gsize==2 && (global.log_span1 || !exspan)) {
                            cerr << "Span: 0[" << g0.gsize << "," << range0 << "]";
                            for (int k = 0; k < g0.gsize; k++)
                                cerr << " " << (g0.pgene[k]==0 ? "0" : g0.pgene[k]->name);
                            cerr << " - " << (exspan ? "exceeded" : "ok") << endl;
                        }
                        if (exspan) continue;
                        SetBestHit(g0, 1);      // set E & ?BH for Y

                        if (g0.gsize > 2) {     // Z required
                            if (g0.pgene[2] == 0) {     // Select gene Z
                                Giter zit = yit;
                                for (++zit; zit != x.cit->second.genes.end(); zit++) {
                                    if (&(zit->second) == g0.pgene[0]) continue;
                                    g0.pgene[2] = &(zit->second);
                                    exspan = ExceedSpan(g0, range0, 0);    // Check span
                                    if (logging && (global.log_span1 || !exspan)) {
                                        cerr << "Span: 0[" << g0.gsize << "," << range0 << "]";
                                        for (int k = 0; k < g0.gsize; k++)
                                            cerr << " " << (g0.pgene[k]==0 ? "0" : g0.pgene[k]->name);
                                        cerr << " - " << (exspan ? "exceeded" : "ok") << endl;
                                    }
                                    if (exspan) continue;
                                    SetBestHit(g0, 2);       // set E & ?BH for Z

                                    insynt = HasSynteny(hsmode, sp1, t.range, subrange);
                                    if (logging && (global.log_synteny || insynt)) {
                                        cerr << "Synteny: 0[" << g0.gsize << "]";
                                        for (int k = 0; k < g0.gsize; k++)
                                            cerr << " " << (g0.pgene[k]==0 ? "0" : g0.pgene[k]->name);
                                        cerr << " 1[" << g1.gsize << "," << t.range << "]";
                                        for (int k = 0; k < g1.gsize; k++)
                                            cerr << " " << (g1.pgene[k]==0 ? "0" : g1.pgene[k]->name);
                                        cerr << " - " << (insynt ? "yes" : "no") << endl;
                                    }
                                    if (insynt) { 
                                        incase = 1; break;      // exit Z-loop
                                    }
                                }       // next Z
                                if (!insynt) g0.pgene[2] = 0;
                            }
                            else {                      // Use fixed Z
                                insynt = HasSynteny(hsmode, sp1, t.range, subrange);
                                if (logging && (global.log_synteny || insynt)) {
                                    cerr << "Synteny: 0[" << g0.gsize << "]";
                                    for (int k = 0; k < g0.gsize; k++)
                                        cerr << " " << (g0.pgene[k]==0 ? "0" : g0.pgene[k]->name);
                                    cerr << " 1[" << g1.gsize << "," << t.range << "]";
                                    for (int k = 0; k < g1.gsize; k++)
                                        cerr << " " << (g1.pgene[k]==0 ? "0" : g1.pgene[k]->name);
                                    cerr << " - " << (insynt ? "yes" : "no") << endl;
                                }
                                if (insynt) incase = 2;
                            }
                        }
                        else {                  // only X,Y required
                            g0.pgene[2] = 0;
                            insynt = HasSynteny(hsmode, sp1, t.range, subrange);
                            if (logging && (global.log_synteny || insynt)) {
                                cerr << "Synteny: 0[" << g0.gsize << "]";
                                for (int k = 0; k < g0.gsize; k++)
                                    cerr << " " << (g0.pgene[k]==0 ? "0" : g0.pgene[k]->name);
                                cerr << " 1[" << g1.gsize << "," << t.range << "]";
                                for (int k = 0; k < g1.gsize; k++)
                                    cerr << " " << (g1.pgene[k]==0 ? "0" : g1.pgene[k]->name);
                                cerr << " - " << (insynt ? "yes" : "no") << endl;
                            }
                            if (insynt) incase = 3;
                        }
                        if (insynt) break;  // exit Y-loop
                    }       // next Y
                    if (!insynt) g0.pgene[1] = 0;
                }
                else {                      // Use fixed Y
                    if (g0.gsize > 2) {
                        if (g0.pgene[2] == 0) {     // Select gene Z
                            Giter zit = x.cit->second.genes.begin();
                            for ( ; zit != x.cit->second.genes.end(); zit++) {
                                if (&(zit->second) == g0.pgene[0]) continue;
                                if (&(zit->second) == g0.pgene[1]) continue;
                                g0.pgene[2] = &(zit->second);
                                exspan = ExceedSpan(g0, range0, 0);    // Check span
                                if (logging && (global.log_span1 || !exspan)) {
                                    cerr << "Span: 0[" << g0.gsize << "," << range0 << "]";
                                    for (int k = 0; k < g0.gsize; k++)
                                        cerr << " " << (g0.pgene[k]==0 ? "0" : g0.pgene[k]->name);
                                    cerr << " - " << (exspan ? "exceeded" : "ok") << endl;
                                }
                                if (exspan) continue;
                                SetBestHit(g0, 2);       // set E & ?BH for Z

                                insynt = HasSynteny(hsmode, sp1, t.range, subrange);
                                if (logging && (global.log_synteny || insynt)) {
                                    cerr << "Synteny: 0[" << g0.gsize << "]";
                                    for (int k = 0; k < g0.gsize; k++)
                                        cerr << " " << (g0.pgene[k]==0 ? "0" : g0.pgene[k]->name);
                                    cerr << " 1[" << g1.gsize << "," << t.range << "]";
                                    for (int k = 0; k < g1.gsize; k++)
                                        cerr << " " << (g1.pgene[k]==0 ? "0" : g1.pgene[k]->name);
                                    cerr << " - " << (insynt ? "yes" : "no") << endl;
                                }
                                if (insynt) { 
                                    incase = 4; break;      // Z-loop
                                }
                            }       // next Z
                        }
                        else {                      // Use fixed Z 
                            insynt = HasSynteny(hsmode, sp1, t.range, subrange);
                            if (logging && (global.log_synteny || insynt)) {
                                cerr << "Synteny: 0[" << g0.gsize << "]";
                                for (int k = 0; k < g0.gsize; k++)
                                    cerr << " " << (g0.pgene[k]==0 ? "0" : g0.pgene[k]->name);
                                cerr << " 1[" << g1.gsize << "," << t.range << "]";
                                for (int k = 0; k < g1.gsize; k++)
                                    cerr << " " << (g1.pgene[k]==0 ? "0" : g1.pgene[k]->name);
                                cerr << " - " << (insynt ? "yes" : "no") << endl;
                            }
                            if (insynt) incase = 5;
                        }
                    }       // endif (g0.gsize > 2)
                    else {      // only X,Y required
                        g0.pgene[2] = 0;
                        insynt = HasSynteny(hsmode, sp1, t.range, subrange);
                        if (logging && (global.log_synteny || insynt)) {
                            cerr << "Synteny: 0[" << g0.gsize << "]";
                            for (int k = 0; k < g0.gsize; k++)
                                cerr << " " << (g0.pgene[k]==0 ? "0" : g0.pgene[k]->name);
                            cerr << " 1[" << g1.gsize << "," << t.range << "]";
                            for (int k = 0; k < g1.gsize; k++)
                                cerr << " " << (g1.pgene[k]==0 ? "0" : g1.pgene[k]->name);
                            cerr << " - " << (insynt ? "yes" : "no") << endl;
                        }
                        if (insynt) incase = 6;
                    }
                }
            }       // endif (g0.gsize > 1)
            else {      // only X required
                g0.pgene[1] = 0; g0.pgene[2] = 0;
                insynt = HasSynteny(hsmode, sp1, t.range, subrange);
                if (logging && (global.log_synteny || insynt)) {
                    cerr << "Synteny: 0[" << g0.gsize << "]";
                    for (int k = 0; k < g0.gsize; k++)
                        cerr << " " << (g0.pgene[k]==0 ? "0" : g0.pgene[k]->name);
                    cerr << " 1[" << g1.gsize << "," << t.range << "]";
                    for (int k = 0; k < g1.gsize; k++)
                        cerr << " " << (g1.pgene[k]==0 ? "0" : g1.pgene[k]->name);
                    cerr << " - " << (insynt ? "yes" : "no") << endl;
                }
                if (insynt) incase = 7;
            }

            if (logging) {
                if (insynt) {
                    cerr << "+++ Synteny #" << incase << " found: 0[" << g0.gsize << "," << range0 << "]";
                    for (int k = 0; k < g0.gsize; k++)
                        cerr << " " << (g0.pgene[k]==0 ? "0" : g0.pgene[k]->name);
                    cerr << " 1[" << g1.gsize << "," << t.range << "]";
                    for (int k = 0; k < g1.gsize; k++)
                        cerr << " " << (g1.pgene[k]==0 ? "0" : g1.pgene[k]->name);
                    cerr << endl;
                }
                else
                    cerr << "--- No synteny found." << endl;
            }
            if (!insynt) {
                g1.clear(); g1.gsize = g0.gsize;
            }
            tno = insynt ? t.yes : t.no;
            if (orcount >=0 && insynt) {
                orcount++;
                if (logging)
                    cerr << "OR_count=" << orcount << endl;
            }
            if (andcount >=0 && !insynt) {
                andcount++;
                if (logging)
                    cerr << "AND_count=" << andcount << endl;
            }
            incase = 0;
            insynt = false;
        }
        else if (t.oper == OVER || t.oper == OVERBK) {
            int hsmode = t.oper == OVERBK ? 1 : 0;
            bool insynt = false;
            subrange = 0;
            int incase = 0;     // synteny case no. for logging
            if (overloop == 0) {    // Initialize loop variables
                g0.gsize = t.nwit + 1;
                for (int n=0; n<maxgroup; n++) {
                    if ((t.uselast0 & (1<<n)) == 0) {
                        g0.pgene[n] = 0;
                        g0.goon |= (1<<n);
                    }
                }
                SetBestHit(g0, 0);       // set E & ?BH for X
                y_no = 0;
                oyit = x.cit->second.genes.begin();
                ozit = oyit; ozit++;
                g1.clear();
                g1.gsize = t.nwit + 1;
                for (int n=0; n<maxgroup; n++)
                    g1.goon |= (1<<n);
            }
            else {              // Increment loop variables
            }

            // Milestones 
            if (global.mpirank == 0 && global.currentgenestep && y_no % global.currentgenestep == 0) {
                ostringstream oss;
                oss << left << "c" << setw(5) << global.c_no << "x" << setw(6) 
                    << global.x_no << "o" << setw(6) << overloop << "y" << setw(6) 
                    << y_no << setw(10) << " ";
                if (logany)     cout << oss.str() << "\r";
                else            cerr << oss.str() << "\r";
            }

            overloop++;
            sp1 = t.sp;
            range1 = t.range;
            // Outer loop until end
            if (g0.gsize>1) {     // Y required
                if ((t.uselast0 & 2) == 0) {    // Select gene Y
                    for ( ; oyit != x.cit->second.genes.end(); oyit++,y_no++) {
                        if (&(oyit->second) == g0.pgene[0]) continue;
                        g0.pgene[1] = &(oyit->second);
                        exspan = ExceedSpan(g0, range0, 0);    // Check span
                        if (logging && g0.gsize==2 && (global.log_span1 || !exspan)) {
                            cerr << "Span: 0[" << g0.gsize << "," << range0 << "]";
                            for (int k = 0; k < g0.gsize; k++)
                                cerr << " " << (g0.pgene[k]==0 ? "0" : g0.pgene[k]->name);
                            cerr << " - " << (exspan ? "exceeded" : "ok") << endl;
                        }
                        if (exspan) continue;
                        SetBestHit(g0, 1);       // set E & ?BH for Y

                        if (g0.gsize > 2) {     // Z required
                            if ((t.uselast0 & 4) == 0) {    // Select gene Z
                                for ( ; ozit != x.cit->second.genes.end(); ozit++) {
                                    if (&(ozit->second) == g0.pgene[0]) continue;
                                    if (&(ozit->second) == g0.pgene[1]) continue;
                                    g0.pgene[2] = &(ozit->second);
                                    exspan = ExceedSpan(g0, range0, 0);    // Check span
                                    if (logging && (global.log_span1 || !exspan)) {
                                        cerr << "Span: 0[" << g0.gsize << "," << range0 << "]";
                                        for (int k = 0; k < g0.gsize; k++)
                                            cerr << " " << (g0.pgene[k]==0 ? "0" : g0.pgene[k]->name);
                                        cerr << " - " << (exspan ? "exceeded" : "ok") << endl;
                                    }
                                    if (exspan) continue;
                                    SetBestHit(g0, 2);       // set E & ?BH for Z

                                    insynt = HasSynteny(hsmode, sp1, t.range, subrange);
                                    if (logging && (global.log_synteny || insynt)) {
                                        cerr << "Synteny: 0[" << g0.gsize << "]";
                                        for (int k = 0; k < g0.gsize; k++)
                                            cerr << " " << (g0.pgene[k]==0 ? "0" : g0.pgene[k]->name);
                                        cerr << " 1[" << g1.gsize << "," << t.range << "]";
                                        for (int k = 0; k < g1.gsize; k++)
                                            cerr << " " << (g1.pgene[k]==0 ? "0" : g1.pgene[k]->name);
                                        cerr << " - " << (insynt ? "yes" : "no") << endl;
                                    }
                                    if (insynt) { 
                                        incase = 1;
                                        break;      // exit Z-loop
                                    }
                                }       // next Z
                                if (!insynt) {
                                    g0.pgene[2] = 0;
                                    ozit = x.cit->second.genes.begin();
                                }
                            }
                            else {              // Use fixed Z 
                                insynt = HasSynteny(hsmode, sp1, t.range, subrange);
                                if (logging && (global.log_synteny || insynt)) {
                                    cerr << "Synteny: 0[" << g0.gsize << "]";
                                    for (int k = 0; k < g0.gsize; k++)
                                        cerr << " " << (g0.pgene[k]==0 ? "0" : g0.pgene[k]->name);
                                    cerr << " 1[" << g1.gsize << "," << t.range << "]";
                                    for (int k = 0; k < g1.gsize; k++)
                                        cerr << " " << (g1.pgene[k]==0 ? "0" : g1.pgene[k]->name);
                                    cerr << " - " << (insynt ? "yes" : "no") << endl;
                                }
                                if (insynt) incase = 2;
                            }
                        }
                        else {      // only X,Y required
                            insynt = HasSynteny(hsmode, sp1, t.range, subrange);
                            if (logging && (global.log_synteny || insynt)) {
                                cerr << "Synteny: 0[" << g0.gsize << "]";
                                for (int k = 0; k < g0.gsize; k++)
                                    cerr << " " << (g0.pgene[k]==0 ? "0" : g0.pgene[k]->name);
                                cerr << " 1[" << g1.gsize << "," << t.range << "]";
                                for (int k = 0; k < g1.gsize; k++)
                                    cerr << " " << (g1.pgene[k]==0 ? "0" : g1.pgene[k]->name);
                                cerr << " - " << (insynt ? "yes" : "no") << endl;
                            }
                            if (insynt) incase = 3;
                        }
                        if (insynt) break;      // Y-loop
                    }       // next Y
                    if (!insynt) {
                        g0.pgene[1] = 0;
                        oyit = x.cit->second.genes.begin();
                    }
                }
                else {                      // Use fixed Y
                    if (g0.gsize > 2) {
                        if ((t.uselast0 & 4) == 0) {    // Select gene Z
                            for ( ; ozit != x.cit->second.genes.end(); ozit++) {
                                if (&(ozit->second) == g0.pgene[0]) continue;
                                if (&(ozit->second) == g0.pgene[1]) continue;
                                g0.pgene[2] = &(ozit->second);
                                exspan = ExceedSpan(g0, range0, 0);    // Check span
                                if (logging && (global.log_span1 || !exspan)) {
                                    cerr << "Span: 0[" << g0.gsize << "]";
                                    for (int k = 0; k < g0.gsize; k++)
                                        cerr << " " << (g0.pgene[k]==0 ? "0" : g0.pgene[k]->name);
                                    cerr << " - " << (exspan ? "exceeded" : "ok") << endl;
                                }
                                if (exspan) continue;
                                SetBestHit(g0, 2);       // set E & ?BH for Z

                                insynt = HasSynteny(hsmode, sp1, t.range, subrange);
                                if (logging && (global.log_synteny || insynt)) {
                                    cerr << "Synteny: 0[" << g0.gsize << "]";
                                    for (int k = 0; k < g0.gsize; k++)
                                        cerr << " " << (g0.pgene[k]==0 ? "0" : g0.pgene[k]->name);
                                    cerr << " 1[" << g1.gsize << "," << t.range << "]";
                                    for (int k = 0; k < g1.gsize; k++)
                                        cerr << " " << (g1.pgene[k]==0 ? "0" : g1.pgene[k]->name);
                                    cerr << " - " << (insynt ? "yes" : "no") << endl;
                                }
                                if (insynt) { 
                                    incase = 4; break;      // Z-loop
                                }
                            }       // next Z
                            if (!insynt) {
                                g0.pgene[2] = 0;
                                ozit = x.cit->second.genes.begin();
                            }
                        }
                        else {                          // Use fixed Z 
                            insynt = HasSynteny(hsmode, sp1, t.range, subrange);
                            if (logging && (global.log_synteny || insynt)) {
                                cerr << "Synteny: 0[" << g0.gsize << "]";
                                for (int k = 0; k < g0.gsize; k++)
                                    cerr << " " << (g0.pgene[k]==0 ? "0" : g0.pgene[k]->name);
                                cerr << " 1[" << g1.gsize << "," << t.range << "]";
                                for (int k = 0; k < g1.gsize; k++)
                                    cerr << " " << (g1.pgene[k]==0 ? "0" : g1.pgene[k]->name);
                                cerr << " - " << (insynt ? "yes" : "no") << endl;
                            }
                            if (insynt) incase = 5;
                        }
                    }       // endif (g0.gsize > 2)
                    else {      // only X,Y required
                        g0.pgene[2] = 0;
                        insynt = HasSynteny(hsmode, sp1, t.range, subrange);
                        if (logging && (global.log_synteny || insynt)) {
                            cerr << "Synteny: 0[" << g0.gsize << "]";
                            for (int k = 0; k < g0.gsize; k++)
                                cerr << " " << (g0.pgene[k]==0 ? "0" : g0.pgene[k]->name);
                            cerr << " 1[" << g1.gsize << "," << t.range << "]";
                            for (int k = 0; k < g1.gsize; k++)
                                cerr << " " << (g1.pgene[k]==0 ? "0" : g1.pgene[k]->name);
                            cerr << " - " << (insynt ? "yes" : "no") << endl;
                        }
                        if (insynt) incase = 6;
                    }
                }
            }       // endif (g0.gsize > 1)
            else {      // only X required
                g0.pgene[1] = 0; g0.pgene[2] = 0;
                insynt = HasSynteny(hsmode, sp1, t.range, subrange);
                if (logging && (global.log_synteny || insynt)) {
                    cerr << "Synteny: 0[" << g0.gsize << "]";
                    for (int k = 0; k < g0.gsize; k++)
                        cerr << " " << (g0.pgene[k]==0 ? "0" : g0.pgene[k]->name);
                    cerr << " 1[" << g1.gsize << "," << t.range << "]";
                    for (int k = 0; k < g1.gsize; k++)
                        cerr << " " << (g1.pgene[k]==0 ? "0" : g1.pgene[k]->name);
                    cerr << " - " << (insynt ? "yes" : "no") << endl;
                }
                if (insynt) incase = 7;
            }

            if (logging) {
                if (insynt) {
                    cerr << "+++ Synteny #" << incase << " found: 0[" << g0.gsize << "," << range0 << "]";
                    for (int k = 0; k < g0.gsize; k++)
                        cerr << " " << (g0.pgene[k]==0 ? "0" : g0.pgene[k]->name);
                    cerr << " 1[" << g1.gsize << "," << t.range << "]";
                    for (int k = 0; k < g1.gsize; k++)
                        cerr << " " << (g1.pgene[k]==0 ? "0" : g1.pgene[k]->name);
                    cerr << endl;
                }
                else
                    cerr << "--- No synteny found." << endl;
            }
            if (insynt) {
                // Store variables of the loop
                tno = t.yes;
            }
            else {
                if (overloop > 0) {     // End of outer loop
                    overloop = 0;
                    if (orcount >= 0 && insynt) {
                        orcount++;
                        if (logging)
                            cerr << "OR_count=" << orcount << endl;
                    }
                    if (andcount >= 0 && !insynt) {
                        andcount++;
                        if (logging)
                            cerr << "AND_count=" << andcount << endl;
                    }
                }
                tno = t.no;
                g1.clear(); g1.gsize = g0.gsize;
                g0.goon = 0;
            }
            incase = 0;
            insynt = false;
        }
        else if (t.oper == IN2 || t.oper == IN2BK) {
            Gene & x2 = *(g1.pgene[0]);     // g2={X",Y",Z"}
            int hsmode = t.oper == IN2BK ? 3 : 2;
            bool insynt = false;
            subrange = 0;
            int incase = 0;     // synteny case no. for logging
            g2 = g1;
            for (int n=0; n<maxgroup; n++) {    // use only needed witnesses from sp1
                if ((t.uselast1 & (1<<n)) == 0)
                    g2.pgene[n] = 0;
            }
            g2.gsize = t.nwit + 1;
            g2.goon = 0;
            SetBestHit(g2, 0);       // set E & ?BH for X
            sp2 = g1.pgene[0]->sp;
            sp3 = t.sp;
            g3.clear();
            g3.gsize = t.nwit + 1;
            // Inner loop until satisfied
            if (g2.gsize > 1) {     // Y" required
                if (g2.pgene[1] == 0) {     // Select gene Y"
                    Giter yit = x2.cit->second.genes.begin();
                    for ( ; yit != x2.cit->second.genes.end(); yit++) {
                        if (&(yit->second) == g2.pgene[0]) continue;
                        g2.pgene[1] = &(yit->second);
                        exspan = ExceedSpan(g2, range1, subrange);    // Check span
                        if (logging && g2.gsize==2 && (global.log_span1 || !exspan)) {
                            cerr << "Span: 2[" << g2.gsize << "," << range1 << "]";
                            for (int k = 0; k < g2.gsize; k++)
                                cerr << " " << (g2.pgene[k]==0 ? "0" : g2.pgene[k]->name);
                            cerr << " - " << (exspan ? "exceeded" : "ok") << endl;
                        }
                        if (exspan) continue;
                        SetBestHit(g2, 1);       // set E & ?BH for Y"

                        if (g2.gsize > 2) {     // Z" required
                            if (g2.pgene[2] == 0) {     // Select gene Z"
                                Giter zit = yit;
                                for (++zit; zit != x2.cit->second.genes.end(); zit++) {
                                    if (&(zit->second) == g2.pgene[0]) continue;
                                    g2.pgene[2] = &(zit->second);
                                    exspan = ExceedSpan(g2, range1, subrange);    // Check span
                                    if (logging && (global.log_span1 || !exspan)) {
                                        cerr << "Span: 2[" << g2.gsize << "," << range1 << "]";
                                        for (int k = 0; k < g2.gsize; k++)
                                            cerr << " " << (g2.pgene[k]==0 ? "0" : g2.pgene[k]->name);
                                        cerr << " - " << (exspan ? "exceeded" : "ok") << endl;
                                    }
                                    if (exspan) continue;
                                    SetBestHit(g2, 2);       // set E & ?BH for Z"

                                    insynt = HasSynteny(hsmode, sp3, t.range, subrange);
                                    if (logging && (global.log_synteny || insynt)) {
                                        cerr << "Synteny: 2[" << g2.gsize << "]";
                                        for (int k = 0; k < g2.gsize; k++)
                                            cerr << " " << (g2.pgene[k]==0 ? "0" : g2.pgene[k]->name);
                                        cerr << " 3[" << g3.gsize << "," << t.range << "]";
                                        for (int k = 0; k < g3.gsize; k++)
                                            cerr << " " << (g3.pgene[k]==0 ? "0" : g3.pgene[k]->name);
                                        cerr << " - " << (insynt ? "yes" : "no") << endl;
                                    }
                                    if (insynt) {
                                        incase = 1; break;      // exit Z-loop
                                    }
                                }       // next Z"
                                if (!insynt) g2.pgene[2] = 0;
                            }
                            else {                      // Use fixed Z"
                                insynt = HasSynteny(hsmode, sp3, t.range, subrange);
                                if (logging && (global.log_synteny || insynt)) {
                                    cerr << "Synteny: 2[" << g2.gsize << "]";
                                    for (int k = 0; k < g2.gsize; k++)
                                        cerr << " " << (g2.pgene[k]==0 ? "0" : g2.pgene[k]->name);
                                    cerr << " 3[" << g3.gsize << "," << t.range << "]";
                                    for (int k = 0; k < g3.gsize; k++)
                                        cerr << " " << (g3.pgene[k]==0 ? "0" : g3.pgene[k]->name);
                                    cerr << " - " << (insynt ? "yes" : "no") << endl;
                                }
                                if (insynt) incase = 2;
                            }
                        }
                        else {                  // only X",Y" required
                            g2.pgene[2] = 0;
                            insynt = HasSynteny(hsmode, sp3, t.range, subrange);
                            if (logging && (global.log_synteny || insynt)) {
                                cerr << "Synteny: 2[" << g2.gsize << "]";
                                for (int k = 0; k < g2.gsize; k++)
                                    cerr << " " << (g2.pgene[k]==0 ? "0" : g2.pgene[k]->name);
                                cerr << " 3[" << g3.gsize << "," << t.range << "]";
                                for (int k = 0; k < g3.gsize; k++)
                                    cerr << " " << (g3.pgene[k]==0 ? "0" : g3.pgene[k]->name);
                                cerr << " - " << (insynt ? "yes" : "no") << endl;
                            }
                            if (insynt) incase = 3;
                        }
                        if (insynt) break;  // exit Y"-loop
                    }       // next Y"
                    if (!insynt) g2.pgene[1] = 0;
                }
                else {                      // Use fixed Y"
                    if (g2.gsize > 2) {
                        if (g2.pgene[2] == 0) {     // Select gene Z"
                            Giter zit = x2.cit->second.genes.begin();
                            for ( ; zit != x2.cit->second.genes.end(); zit++) {
                                if (&(zit->second) == g2.pgene[0]) continue;
                                if (&(zit->second) == g2.pgene[1]) continue;
                                g2.pgene[2] = &(zit->second);
                                exspan = ExceedSpan(g2, range1, subrange);    // Check span
                                if (logging && (global.log_span1 || !exspan)) {
                                    cerr << "Span: 2[" << g2.gsize << "," << range1 << "]";
                                    for (int k = 0; k < g2.gsize; k++)
                                        cerr << " " << (g2.pgene[k]==0 ? "0" : g2.pgene[k]->name);
                                    cerr << " - " << (exspan ? "exceeded" : "ok") << endl;
                                }
                                if (exspan) continue;
                                SetBestHit(g2, 2);       // set E & ?BH for Z"

                                insynt = HasSynteny(hsmode, sp3, t.range, subrange);
                                if (logging && (global.log_synteny || insynt)) {
                                    cerr << "Synteny: 2[" << g2.gsize << "]";
                                    for (int k = 0; k < g2.gsize; k++)
                                        cerr << " " << (g2.pgene[k]==0 ? "0" : g2.pgene[k]->name);
                                    cerr << " 3[" << g3.gsize << "," << t.range << "]";
                                    for (int k = 0; k < g3.gsize; k++)
                                        cerr << " " << (g3.pgene[k]==0 ? "0" : g3.pgene[k]->name);
                                    cerr << " - " << (insynt ? "yes" : "no") << endl;
                                }
                                if (insynt) {
                                    incase = 4; break;
                                }
                            }       // next Z"
                        }
                        else {                      // Use fixed Z"
                            insynt = HasSynteny(hsmode, sp3, t.range, subrange);
                            if (logging && (global.log_synteny || insynt)) {
                                cerr << "Synteny: 2[" << g2.gsize << "]";
                                for (int k = 0; k < g2.gsize; k++)
                                    cerr << " " << (g2.pgene[k]==0 ? "0" : g2.pgene[k]->name);
                                cerr << " 3[" << g3.gsize << "," << t.range << "]";
                                for (int k = 0; k < g3.gsize; k++)
                                    cerr << " " << (g3.pgene[k]==0 ? "0" : g3.pgene[k]->name);
                                cerr << " - " << (insynt ? "yes" : "no") << endl;
                            }
                            if (insynt) incase = 5;
                        }
                    }       // endif (g2.gsixe > 2)
                    else {      // only X",Y" required
                        g2.pgene[2] = 0;
                        insynt = HasSynteny(hsmode, sp3, t.range, subrange);
                        if (logging && (global.log_synteny || insynt)) {
                            cerr << "Synteny: 2[" << g2.gsize << "]";
                            for (int k = 0; k < g2.gsize; k++)
                                cerr << " " << (g2.pgene[k]==0 ? "0" : g2.pgene[k]->name);
                            cerr << " 3[" << g3.gsize << "," << t.range << "]";
                            for (int k = 0; k < g3.gsize; k++)
                                cerr << " " << (g3.pgene[k]==0 ? "0" : g3.pgene[k]->name);
                            cerr << " - " << (insynt ? "yes" : "no") << endl;
                        }
                        if (insynt) incase = 6;
                    }
                }
            }       // endif (g2.gsize > 1)
            else {      // only X" required
                g2.pgene[1] = 0; g2.pgene[2] = 0;
                insynt = HasSynteny(hsmode, sp3, t.range, subrange);
                if (logging && (global.log_synteny || insynt)) {
                    cerr << "Synteny: 2[" << g2.gsize << "]";
                    for (int k = 0; k < g2.gsize; k++)
                        cerr << " " << (g2.pgene[k]==0 ? "0" : g2.pgene[k]->name);
                    cerr << " 3[" << g3.gsize << "," << t.range << "]";
                    for (int k = 0; k < g3.gsize; k++)
                        cerr << " " << (g3.pgene[k]==0 ? "0" : g3.pgene[k]->name);
                    cerr << " - " << (insynt ? "yes" : "no") << endl;
                }
                if (insynt) incase = 7;
            }
            if (logging) {
                if (insynt) {
                    cerr << "+++ Synteny2 #" << incase << " found: 2[" << g2.gsize << "," << range1 << "]";
                    for (int k = 0; k < g2.gsize; k++)
                        cerr << " " << (g2.pgene[k]==0 ? "0" : g2.pgene[k]->name);
                    cerr << " 3[" << g3.gsize << "," << t.range << "]";
                    for (int k = 0; k < g3.gsize; k++)
                        cerr << " " << (g3.pgene[k]==0 ? "0" : g3.pgene[k]->name);
                    cerr << endl;
                }
                else
                    cerr << "--- No synteny2 found." << endl;
            }
            if (!insynt) {
                g3.clear(); g3.gsize = g2.gsize;
            }
            tno = insynt ? t.yes : t.no;
            incase = 0;
            insynt = false;
        }
        else if (t.oper == ADD) {
            if (t.sp == sp0 || t.sp < 0 && global.taxa[-t.sp].specno[0] == sp0) {
                if (logging) 
                    cerr << "Groups[" << groups.size() << "]=Group0" << endl;
                groups.push_back(g0);
                if (global.contin != 0 && ogit != oldgroups.end()) ogit++;
            }
            // allow for sp1==sp2 (3 spp predicate)
            else if (t.sp == sp1 && (g1.pgene[0] == 0 || groups.back().pgene[0] == 0 || 
                                     groups.back().pgene[0]->sp != g1.pgene[0]->sp)) {
                if (logging) 
                    cerr << "Groups[" << groups.size() << "]=Group1" << endl;
                groups.push_back(g1);
                if (global.contin != 0 && ogit != oldgroups.end()) ogit++;
            }
            else if (t.sp == sp2 || t.sp < 0 && t.sp == sp1) {
                if (logging) 
                    cerr << "Groups[" << groups.size() << "]=Group2" << endl;
                groups.push_back(g2);
                if (global.contin != 0 && ogit != oldgroups.end()) ogit++;
            }
            else if (t.sp == sp3) {
                if (logging) 
                    cerr << "Groups[" << groups.size() << "]=Group3" << endl;
                groups.push_back(g3);
                if (global.contin != 0 && ogit != oldgroups.end()) ogit++;
            }
            else {
                ostringstream oss;
                if (global.mpisize > 1) oss << "[" << global.mpirank << "] ";
                oss << "Runtime error: invalid ADD term " << tno;
                cerr << oss.str() << endl;
                return false;
            }
            tno = t.yes;
        }
        else {
            ostringstream oss;
            if (global.mpisize > 1) oss << "[" << global.mpirank << "] ";
            oss << "Runtime error: invalid term " << tno;
            cerr << oss.str() << endl;
            return false;
        }
    }
    if (logging)
        cerr << "Predicate is " << boolalpha << (tno == YES) << endl;
    gt0 = parminit;         // restore initial parameters
    return (tno == YES);
}

// Return true/false, side effects in g0...g3
//  mode  0:ga,gb=g0,g1(IN,OVER)  1:ga,gb=g0,g1(INBK,OVERBK)  2:ga,gb=g2,g3(IN2)  3:ga,gb=g2,g3(IN2BK)
bool HasSynteny(int mode, int specb, const int span, const int subrange) {
    Group & ga = mode < 2 ? g0 : g2; 
    Group & gb = mode < 2 ? g1 : g3;
    bool testback = mode == 1 || mode == 3;
    if (ga.gsize != gb.gsize) return false;
    vector <Homolog>::const_iterator hitx;
    vector <Homolog>::const_iterator hity;
    vector <Homolog>::const_iterator hitz;
    PGene px, py, pz;
    Group grp(gb);
    bool logging = (global.mpisize == 1) && (global.log_verbose || ga.pgene[0]->log);
    bool exspan = false;
    bool insynt = false;
    bool zerop = false;     // zero pointer flag
    int spa = ga.pgene[0]->sp;
    int spb = unknown_species;

    if (gb.pgene[0] == 0) {         // start search
        if (specb >= 0) 
            spb = specb;
        else if (global.taxa[-specb].nspec > 0)
            spb = global.taxa[-specb].specno[0];
        else
            return false;
    }
    else {                          // continue search
        spb = gb.pgene[0]->sp;
    }

    // Do for species spb and optionally over taxon -specb beyond spb
    while (spb != unknown_species) {
        // Loop over X' ~ X
        float sbx1 = 0.0;
        vector <Homolog> hvx( ga.pgene[0]->homologs[spb] );
        if (global.mode_homology && global.mode_info && !(ga.BBH & ORTX)) {
            // add unique paralogs of each ortholog
            vector <Homolog> para;
            vector <Homolog>::iterator xpi = hvx.begin();
            for ( ; xpi != hvx.end(); xpi++) {
                vector <Homolog>& parxp = xpi->pgene->homologs[spb]; // paralogs of ortholog
                vector <Homolog>::iterator pitxp = parxp.begin();
                for ( ; pitxp != parxp.end(); pitxp++) {
                    vector <Homolog>::iterator herex = hvx.begin();
                    for ( ; herex != hvx.end(); herex++)
                        if (pitxp->pgene == herex->pgene) break;
                    if (herex == hvx.end())
                        para.push_back(*pitxp);  // will be added to the list of orthologs
                }
            }
            vector <Homolog>::iterator pi = para.begin();
            for ( ; pi != para.end(); pi++)
                hvx.push_back(*pi);
        }
        if (global.mode_matrix && global.inv_skip == 0 && (ga.BBH & (BBHX | OBHX)))
            sbx1 = GetBestHit(hvx);     // find OBH value
        hitx = hvx.begin();
        if ((grp.goon & 1) && grp.pgene[0]) {   // continue at last X'
            for ( ; hitx != hvx.end(); hitx++) {
                // Check rank, score and E-value
                if (hitx->rank > ga.rank || hitx->score < ga.score || hitx->evalue > ga.E) continue;
                // Check best hit conditions
                if (global.mode_matrix && (ga.BBH & (BBHX | OBHX))) {
                    if (global.inv_skip == 0 && hitx->score < sbx1 ||
                        global.inv_skip == 1 && hitx->rank > 1.0F ||
                        global.inv_skip == 2 && hitx->rank != 0.0F) continue;
                }
                if (global.mode_matrix && (ga.BBH & (BBHX | IBHX))) {
                    if (global.inv_skip == 0) {
                        PGene pgx = hitx->pgene;
                        vector <Homolog> & hgx = pgx->homologs[spa];
                        float sbx2 = GetBestHit(hgx);      // find IBH value
                        vector <Homolog>::iterator hgxit = hgx.begin();
                        for (; hgxit != hgx.end(); hgxit++)
                            if (hgxit->pgene == ga.pgene[0]) break;
                        if (hgxit == hgx.end() || hgxit->score < sbx2) continue;
                    }
                    else if (global.inv_skip == 1 && hitx->brank > 1.0F ||
                        global.inv_skip == 2 && hitx->brank != 0.0F) continue;
                }
                if (hitx->pgene == grp.pgene[0]) break;
            }
            if (hitx == hvx.end()) hitx = hvx.begin();
            else if (ga.gsize == 1 || (global.contin & 1) != 0) hitx++;         // increment X'
        }
        for ( ; hitx != hvx.end(); hitx++) {
            // Check rank, score & E-value
            if (hitx->rank > ga.rank || hitx->score < ga.score || hitx->evalue > ga.E) continue;
            // Check best hit conditions
            if (global.mode_matrix && (ga.BBH & (BBHX | OBHX))) {
                if (global.inv_skip == 0 && hitx->score < sbx1 ||
                    global.inv_skip == 1 && hitx->rank > 1.0F ||
                    global.inv_skip == 2 && hitx->rank != 0.0F) continue;
            }
            if (global.mode_matrix && (ga.BBH & (BBHX | IBHX))) {
                if (global.inv_skip == 0) {
                    PGene pgx = hitx->pgene;
                    vector <Homolog> & hgx = pgx->homologs[spa];
                    float sbx2 = GetBestHit(hgx);      // find IBH value
                    vector <Homolog>::iterator hgxit = hgx.begin();
                    for (; hgxit != hgx.end(); hgxit++)
                        if (hgxit->pgene == ga.pgene[0]) break;
                    if (hgxit == hgx.end() || hgxit->score < sbx2) continue;
                }
                else if (global.inv_skip == 1 && hitx->brank > 1.0F ||
                    global.inv_skip == 2 && hitx->brank != 0.0F) continue;
            }
            px = hitx->pgene;
            grp.pgene[0] = px;
            if (global.mode_matrix) {
                ga.xrank = hitx->rank;
                ga.xscore = hitx->score;
                grp.xrank = hitx->brank;
                grp.xscore = hitx->score;   // this is wrong for asymmetric matrix!
            }

            // Skip contigs with not enough genes
            if ((int)px->cit->second.genes.size() < gb.gsize)
                continue;

            if (ga.gsize > 1) {
                if (ga.pgene[1] == 0) {
                    zerop = true; break;
                }
                // Loop over Y' ~ Y
                float sby1 = 0.0F;
                vector <Homolog> hvy( ga.pgene[1]->homologs[spb] );
                if (global.mode_homology && global.mode_info && !(ga.BBH & ORTY)) {
                    // add unique paralogs of each ortholog
                    vector <Homolog> para;
                    vector <Homolog>::iterator ypi = hvy.begin();
                    for ( ; ypi != hvy.end(); ypi++) {
                        vector <Homolog>& paryp = ypi->pgene->homologs[spb]; // paralogs of ortholog
                        vector <Homolog>::iterator pityp = paryp.begin();
                        for ( ; pityp != paryp.end(); pityp++) {
                            vector <Homolog>::iterator herey = hvy.begin();
                            for ( ; herey != hvy.end(); herey++)
                                if (pityp->pgene == herey->pgene) break;
                            if (herey == hvy.end())
                                para.push_back(*pityp);  // will be added to the list of orthologs
                        }
                    }
                    vector <Homolog>::iterator pi = para.begin();
                    for ( ; pi != para.end(); pi++)
                        hvy.push_back(*pi);
                }
                if (global.mode_matrix && global.inv_skip == 0 && (ga.BBH & (BBHY | OBHY)))
                    sby1 = GetBestHit(hvy);     // find OBH value
                hity = hvy.begin();
                if ((grp.goon & 2) && grp.pgene[1]) {   // continue at last Y'
                    for ( ; hity != hvy.end(); hity++) {
                        // Check rank, score & E-value
                        if (hity->rank > ga.rank || hity->score < ga.score || hity->evalue > ga.E) continue;
                        // Check best hit conditions
                        if (global.mode_matrix && (ga.BBH & (BBHY | OBHY))) {
                            if (global.inv_skip == 0 && hity->score < sby1 ||
                                global.inv_skip == 1 && hity->rank > 1.0F ||
                                global.inv_skip == 2 && hity->rank != 0.0F) continue;
                        }
                        if (global.mode_matrix && (ga.BBH & (BBHY | IBHY))) {
                            if (global.inv_skip == 0) {
                                PGene pgy = hity->pgene;
                                vector <Homolog> & hgy = pgy->homologs[spa];
                                float sby2 = GetBestHit(hgy);      // find IBH value
                                vector <Homolog>::iterator hgyit = hgy.begin();
                                for ( ; hgyit != hgy.end(); hgyit++)
                                    if (hgyit->pgene == ga.pgene[1]) break;
                                if (hgyit == hgy.end() || hgyit->score > sby2) continue;
                            }
                            else if (global.inv_skip == 1 && hity->brank > 1.0F ||
                                global.inv_skip == 2 && hity->brank != 0.0F) continue;
                        }
                        if (hity->pgene == grp.pgene[1]) break;
                    }
                    if (hity == hvy.end()) hity = hvy.begin();
                    else if (ga.gsize == 2) hity++;         // increment Y'
                }
                for ( ; hity != hvy.end(); hity++) {
                    // Check score & E-value
                    if (hity->rank > ga.rank || hity->score < ga.score || hity->evalue > ga.E) continue;
                    // Check different contig
                    if (px->cit != hity->pgene->cit) continue;
                    // Check best hit conditions
                    if (global.mode_matrix && (ga.BBH & (BBHY | OBHY))) {
                        if (global.inv_skip == 0 && hity->score < sby1 ||
                            global.inv_skip == 1 && hity->rank > 1.0F ||
                            global.inv_skip == 2 && hity->rank != 0.0F) continue;
                    }
                    if (global.mode_matrix && (ga.BBH & (BBHY | IBHY))) {
                        if (global.inv_skip == 0) {
                            PGene pgy = hity->pgene;
                            vector <Homolog> & hgy = pgy->homologs[spa];
                            float sby2 = GetBestHit(hgy);      // find IBH value
                            vector <Homolog>::iterator hgyit = hgy.begin();
                            for (; hgyit != hgy.end(); hgyit++)
                                if (hgyit->pgene == ga.pgene[1]) break;
                            if (hgyit == hgy.end() || hgyit->score < sby2) continue;
                        }
                        else if (global.inv_skip == 1 && hity->brank > 1.0F ||
                            global.inv_skip == 2 && hity->brank != 0.0F) continue;
                    }
                    py = hity->pgene;
                    if (py == px) continue;
                    grp.pgene[1] = py;
                 
                    // Check if different species
                    if (px->sp != py->sp) {
                        ostringstream oss;
                        oss << "\n";
                        if (global.mpisize > 1) oss << "[" << global.mpirank << "] ";
                        oss << "Internal error: genes " << px->name << ", " << py->name 
                            << " of diff. spp. " << px->sp << "," << py->sp;
                        cout << oss.str() << endl;
                        //exit(4);
                    }
                 
                    exspan = ExceedSpan(grp, span, subrange);
                    if (logging && ga.gsize==2 && (global.log_span2 || !exspan)) {
                        cerr << "Span: 1[" << grp.gsize << "," << span << "]";
                        for (int k = 0; k < grp.gsize; k++)
                            cerr << " " << (grp.pgene[k]==0 ? "0" : grp.pgene[k]->name);
                        cerr << " - " << (exspan ? "exceeded" : "ok") << endl;
                    }
                    if (exspan) continue;

                    if (ga.gsize > 2) {             // X,Y,Z
                        if (ga.pgene[2] == 0) {
                            zerop = true; break;
                        }
                        // Loop over Z' ~ Z
                        float sbz1 = 0.0;
                        vector <Homolog> hvz( ga.pgene[2]->homologs[spb] );
                        if (global.mode_homology && global.mode_info && !(ga.BBH & ORTZ)) {
                            // add unique paralogs of each ortholog
                            vector <Homolog> para;
                            vector <Homolog>::iterator zpi = hvz.begin();
                            for ( ; zpi != hvz.end(); zpi++) {
                                vector <Homolog>& parzp = zpi->pgene->homologs[spb]; // paralogs of ortholog
                                vector <Homolog>::iterator pitzp = parzp.begin();
                                for ( ; pitzp != parzp.end(); pitzp++) {
                                    vector <Homolog>::iterator herez = hvz.begin();
                                    for ( ; herez != hvz.end(); herez++)
                                        if (pitzp->pgene == herez->pgene) break;
                                    if (herez == hvz.end())
                                        para.push_back(*pitzp);  // will be added to the list of orthologs
                                }
                            }
                            vector <Homolog>::iterator pi = para.begin();
                            for ( ; pi != para.end(); pi++)
                                hvz.push_back(*pi);
                        }
                        if (global.mode_matrix && global.inv_skip == 0 && (ga.BBH & (BBHZ | OBHZ)))
                            sbz1 = GetBestHit(hvz);     // find OBH value
                        hitz = hvz.begin();
                        if ((grp.goon & 4) && grp.pgene[2]) {   // continue at last Z'
                            for ( ; hitz != hvz.end(); hitz++) {
                                // Check rank, score & E-value
                                if (hitz->rank > ga.rank || hitz->score < ga.score || hitz->evalue > ga.E) continue;
                                // Check best hit conditions
                                if (global.mode_matrix && (ga.BBH & (BBHZ | OBHZ))) {
                                    if (global.inv_skip == 0 && hitz->score < sbz1||
                                        global.inv_skip == 1 && hitz->rank > 1.0F ||
                                        global.inv_skip == 2 && hitz->rank != 0.0F) continue;
                                }
                                if (global.mode_matrix && (ga.BBH & (BBHZ | IBHZ))) {
                                    if (global.inv_skip == 0) {
                                        PGene pgz = hitz->pgene;
                                        vector <Homolog> & hgz = pgz->homologs[spa];
                                        float sbz2 = GetBestHit(hgz);      // find IBH value
                                        vector <Homolog>::iterator hgzit = hgz.begin();
                                        for (; hgzit != hgz.end(); hgzit++)
                                            if (hgzit->pgene == ga.pgene[2]) break;
                                        if (hgzit == hgz.end() || hgzit->score < sbz2) continue;
                                    }
                                    else if (global.inv_skip == 1 && hitz->brank > 1.0F ||
                                        global.inv_skip == 2 && hitz->brank != 0.0F) continue;
                                }
                                if (hitz->pgene == grp.pgene[2]) break;
                            }
                            if (hitz == hvz.end()) hitz = hvz.begin();
                            else if (ga.gsize == 3) hitz++;         // increment Z'
                        }
                        for ( ; hitz != hvz.end(); hitz++) {
                            // Check rank, score & E-value
                            if (hitz->rank > ga.rank || hitz->score < ga.score || hitz->evalue > ga.E) continue;
                            // Check different contig
                            if (px->cit != hitz->pgene->cit) continue;
                            // Check best hit conditions
                            if (global.mode_matrix && (ga.BBH & (BBHZ | OBHZ))) {
                                if (global.inv_skip == 0 && hitz->score < sbz1 ||
                                    global.inv_skip == 1 && hitz->rank > 1.0F ||
                                    global.inv_skip == 2 && hitz->rank != 0.0F) continue;
                            }
                            if (global.mode_matrix &&  (ga.BBH & (BBHZ | IBHZ))) {
                                if (global.inv_skip == 0) {
                                    PGene pgz = hitz->pgene;
                                    vector <Homolog> & hgz = pgz->homologs[spa];
                                    float sbz2 = GetBestHit(hgz);      // find IBH value
                                    vector <Homolog>::iterator hgzit = hgz.begin();
                                    for (; hgzit != hgz.end(); hgzit++)
                                        if (hgzit->pgene == ga.pgene[2]) break;
                                    if (hgzit == hgz.end() || hgzit->score < sbz2) continue;
                                }
                                else if (global.inv_skip == 1 && hitz->brank > 1.0F ||
                                    global.inv_skip == 2 && hitz->brank != 0.0F) continue;
                            }
                            pz = hitz->pgene;
                            if (pz == px) continue;
                            if (pz == py && ZneY) continue;
                            grp.pgene[2] = pz;

                            // Check if different species
                            if (px->sp != pz->sp) {
                                ostringstream oss;
                                if (global.mpisize > 1) oss << "[" << global.mpirank << "] ";
                                oss << "Internal error: genes X',Z' from different species " 
                                    << px->sp << "," << pz->sp;
                                cerr << oss.str() << endl;
                                exit(4);
                            }
                  
                            exspan = ExceedSpan(grp, span, subrange);
                            if (logging && (global.log_span2 || !exspan)) {
                                cerr << "Span: 1[" << grp.gsize << "," << span << "]";
                                for (int k = 0; k < grp.gsize; k++)
                                    cerr << " " << (grp.pgene[k]==0 ? "0" : grp.pgene[k]->name);
                                cerr << " - " << (exspan ? "exceeded" : "ok") << endl;
                            }
                            if (exspan) continue;

                            // Check synteny
                            insynt = InSynteny(ga, grp);
                            if (insynt) {
                                gb = grp; break;
                            }
                        }     // next Z'
                        if (!insynt) grp.pgene[2] = 0;
                    }
                    else {                      // only X',Y'
                        grp.pgene[2] = 0;
                        insynt = InSynteny(ga, grp);
                        if (insynt) {
                            gb = grp; break;
                        }
                    }
                }       // next Y'
                if (!insynt) grp.pgene[1] = 0;
                if (insynt || zerop) break;
            }
            else {      // X' only
                gb = grp;
                insynt = true;
                break;
            }
        }       // next X'

        if (!insynt) grp.pgene[0] = 0;
        else if (global.mode_matrix && testback)             // backward checking
            insynt = CheckBack(mode, span);

        if (insynt || specb >= 0) break;    // synteny found

        // try next species from current taxon
        int oldsp2 = spb;
        spb = unknown_species;
        for (int i = 0; i < global.taxa[-specb].nspec; i++) {
            if (global.taxa[-specb].specno[i] == oldsp2) {
                if (++i < global.taxa[-specb].nspec)
                    spb = global.taxa[-specb].specno[i];
                break;
            }
        }
    }
    return insynt;
}

// Return true on successful backward check
bool    CheckBack(int mode, const int span) {
    bool insynt = false;
    Group & ga = mode < 2 ? g0 : g2;
    Group & gb = mode < 2 ? g1 : g3;
    if (ga.gsize == 1) return true;         // nothing to do
    vector <Homolog>::const_iterator hitx;  // homologs of spc in spd
    vector <Homolog>::const_iterator hity;
    vector <Homolog>::const_iterator hitz;
    PGene pcx, pcy, pcz;                    // ptr to genes of spc
    PGene pdx, pdy, pdz;                    // ptr to genes of spd
    pcx = gb.pgene[0];              // X' ptr
    int spc = gb.pgene[0]->sp;
    int spd = ga.pgene[0]->sp;

    // get reliable X'->X score if no ranks and asymmetric matrix
    if (global.inv_skip == 0 && !global.mode_symmetric) {
        vector <Homolog> & hvx = pcx->homologs[spd];
        hitx = hvx.begin();
        for (; hitx != hvx.end(); hitx++) {
            if (hitx->pgene == ga.pgene[0]) {
                gb.xrank = hitx->rank;
                gb.xscore = hitx->score;
                break;
            }
        }
        if (hitx == hvx.end()) {
            gb.xrank = 999.0F;
            gb.xscore = 0.0F;
        }
    }

    // get paralogs of X
    vector <Homolog> & pvx = ga.pgene[0]->homologs[spd];

    Group gc(gb);
    gc.goon = 0;
    for (int n = 1; n < maxgroup; n++) gc.pgene[n] = 0;

    Group gd(ga);       // for alternative (better?) backward homolog with witnesses
    gd.goon = 0;
    for (int n = 0; n < maxgroup; n++) gd.pgene[n] = 0;

    // Loop over U ~ X' excluding ga.pgene[0] (X)
    SetBestHit(gc, 0);       // set E & ?BH for X'
    vector <Homolog> hvx(pcx->homologs[spd]);
    hitx = hvx.begin();
    for (; hitx != hvx.end(); hitx++) {

        // Skip X itself or homologs of X' that are not geater than lambda better of X
        if (hitx->pgene == ga.pgene[0] || hitx->score - gb.xscore <= global.lambda ||           // normalized
            hitx->score > 5.0F && gb.xscore > 5.0F && hitx->score/gb.xscore <= global.lambda)   // unnormalized 
            continue;                           

        // Skip homologs of rank, score & E-value below threshold
        if (hitx->rank > gc.rank || hitx->score < gc.score || hitx->evalue > gc.E) continue;

        pdx = hitx->pgene;
        gd.pgene[0] = pdx;

        // Searching witnesses: start from contig of X', nested loop for Y',Z' in the neighborhood
        Giter yit = pcx->cit->second.genes.begin();
        for (; yit != pcx->cit->second.genes.end(); yit++) {

            pcy = &(yit->second);
            if (pcy == pcx) continue;        // skip X'
            gc.pgene[1] = pcy;

            if (ExceedSpan(gc, span, 0)) continue;
            SetBestHit(gc, 1);       // set E & ?BH for Y'

            // Check X,Y synteny: loop over Y ~ Y'
            float sby = 0.0F;
            vector <Homolog> hvy(pcy->homologs[spd]);
            if (global.inv_skip == 0 && (gc.BBH & (BBHY | OBHY)))
                sby = GetBestHit(hvy);     // find OBH value

            hity = hvy.begin();
            for (; hity != hvy.end(); hity++) {
                // Check score & E-value
                if (hity->rank > gc.rank || hity->score < gc.score || hity->evalue > gc.E) continue;
                // Check different contig
                if (pdx->cit != hity->pgene->cit) continue;
                // Check best hit conditions
                if (gc.BBH & (BBHY | OBHY)) {
                    if (global.inv_skip == 0 && hity->score < sby ||
                        global.inv_skip == 1 && hity->rank > 1.0F ||
                        global.inv_skip == 2 && hity->rank != 0.0F) continue;
                }
                if (gc.BBH & (BBHY | IBHY)) {
                    if (global.inv_skip == 0) {
                        PGene pgy = hity->pgene;
                        vector <Homolog> & hgy = pgy->homologs[spc];
                        float sby2 = GetBestHit(hgy);      // find IBH value
                        vector <Homolog>::iterator hgyit = hgy.begin();
                        for (; hgyit != hgy.end(); hgyit++)
                            if (hgyit->pgene == ga.pgene[1]) break;
                        if (hgyit == hgy.end() || hgyit->score < sby2) continue;
                    }
                    else if (global.inv_skip == 1 && hity->brank > 1.0F ||
                        global.inv_skip == 2 && hity->brank != 0.0F) continue;
                }
                pdy = hity->pgene;
                if (pdy == pdx) continue;
                gd.pgene[1] = pdy;
                if (! ExceedSpan(gd, span, 0)) break;
            }   // next Y

            if (hity == hvy.end()) continue;        // next Y'

            if (gc.gsize > 2) {     // Z required
                Giter zit = yit;
                for (++zit; zit != pcx->cit->second.genes.end(); zit++) {

                    pcz = &(zit->second);
                    if (pcz == pcx) continue;        // skip X'
                    gc.pgene[2] = pcz;

                    if (ExceedSpan(gc, span, 0)) continue;
                    SetBestHit(gc, 2);       // set E & ?BH for Z'

                    // Check X,Y,Z synteny: loop over Z ~ Z'
                    float sbz = 0.0F;
                    vector <Homolog> hvz(pcz->homologs[spd]);
                    if (global.inv_skip == 0 && (gc.BBH & (BBHY | OBHY)))
                        sbz = GetBestHit(hvz);     // find OBH value

                    hitz = hvz.begin();
                    for (; hitz != hvz.end(); hitz++) {
                        // Check score & E-value
                        if (hitz->rank > gc.rank || hitz->score < gc.score || hitz->evalue > gc.E) continue;
                        // Check different contig
                        if (pdx->cit != hitz->pgene->cit) continue;
                        // Check best hit conditions
                        if (gc.BBH & (BBHY | OBHY)) {
                            if (global.inv_skip == 0 && hitz->score < sbz ||
                                global.inv_skip == 1 && hitz->rank > 1.0F ||
                                global.inv_skip == 2 && hitz->rank != 0.0F) continue;
                        }
                        if (gc.BBH & (BBHY | IBHY)) {
                            if (global.inv_skip == 0) {
                                PGene pgz = hitz->pgene;
                                vector <Homolog> & hgz = pgz->homologs[spc];
                                float sbz2 = GetBestHit(hgz);      // find IBH value
                                vector <Homolog>::iterator hgzit = hgz.begin();
                                for (; hgzit != hgz.end(); hgzit++)
                                    if (hgzit->pgene == ga.pgene[2]) break;
                                if (hgzit == hgz.end() || hgzit->score < sbz2) continue;
                            }
                            else if (global.inv_skip == 1 && hitz->brank > 1.0F ||
                                global.inv_skip == 2 && hitz->brank != 0.0F) continue;
                        }
                        pdz = hitz->pgene;
                        if (pdz == pdx) continue;
                        if (pdz == pdy && ZneY) continue;
                        gd.pgene[2] = pdz;
                        if (ExceedSpan(gd, span, 0)) continue;

                        insynt = InSynteny(gc, gd);
                        if (insynt) break;
                    }   // next Z

                    if (insynt) break;
                }   // next Z'
            }   // Z required

            else {      // only X',Y'
                insynt = InSynteny(gc, gd);
                if (insynt) break;
            }
        }   // next Y'
    }
    return ! insynt;
}

// Set rank, score, E-value and BBH flags of <geneno>-th gene in group from current term 
void    SetBestHit(Group & group, int geneno) {
    Term & term = global.terms[tno];
	group.rank = term.rank;
	group.score = term.score;
    group.E = term.E;
    switch (geneno) {
        case 0:
            group.BBH &= ~(BBHX | OBHX | IBHX);
            group.BBH |= term.BBH & (BBHX | OBHX | IBHX);
            break;
        case 1:
            group.BBH &= ~(BBHY | OBHY | IBHY);
            group.BBH |= term.BBH & (BBHY | OBHY | IBHY);
            break;
        case 2:
            group.BBH &= ~(BBHZ | OBHZ | IBHZ);
            group.BBH |= term.BBH & (BBHZ | OBHZ | IBHZ);
            break;
        default:
            if (global.mpirank == 0)
                cerr << "Group of more than 3 genes has not been implemented yet." << endl;
    }
}

// Return best hit value for given list of homologies
float  GetBestHit(vector <Homolog> & hv) {
    float ret = 0;
    vector <Homolog>::iterator hvit = hv.begin();
    for ( ; hvit != hv.end(); hvit++) {
        if (hvit->score > ret)
            ret = hvit->score;
    }
    return ret;
}
